#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2018 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using HeuristicLab.Analysis;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;
using HeuristicLab.Problems.DataAnalysis.Symbolic;
using HeuristicLab.Problems.DataAnalysis.Symbolic.Regression;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Random;

namespace HeuristicLab.Algorithms.DataAnalysis {
  /// <summary>
  /// Bayesian non-linear regression data analysis algorithm.
  /// </summary>
  [Item("Bayesian Nonlinear Regression (BNLR)", "Nonlinear regression algorithm which uses HMC to create samples for the posterior distribution for the model parameters.")]
  [Creatable(CreatableAttribute.Categories.DataAnalysisRegression, Priority = 120)]
  [StorableClass]
  public sealed class BayesianNonlinearRegression : FixedDataAnalysisAlgorithm<IRegressionProblem> {
    private const string RegressionSolutionResultName = "Regression solution";
    private const string ModelStructureParameterName = "Model structure";
    private const string IterationsParameterName = "Iterations";
    private const string RestartsParameterName = "Restarts";
    private const string SetSeedRandomlyParameterName = "SetSeedRandomly";
    private const string SeedParameterName = "Seed";
    private const string InitParamsRandomlyParameterName = "InitializeParametersRandomly";

    public IFixedValueParameter<StringValue> ModelStructureParameter {
      get { return (IFixedValueParameter<StringValue>)Parameters[ModelStructureParameterName]; }
    }
    public IFixedValueParameter<IntValue> IterationsParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters[IterationsParameterName]; }
    }

    public IFixedValueParameter<BoolValue> SetSeedRandomlyParameter {
      get { return (IFixedValueParameter<BoolValue>)Parameters[SetSeedRandomlyParameterName]; }
    }

    public IFixedValueParameter<IntValue> SeedParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters[SeedParameterName]; }
    }

    public IFixedValueParameter<IntValue> RestartsParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters[RestartsParameterName]; }
    }

    public IFixedValueParameter<BoolValue> InitParametersRandomlyParameter {
      get { return (IFixedValueParameter<BoolValue>)Parameters[InitParamsRandomlyParameterName]; }
    }

    public IFixedValueParameter<IntValue> LeapFrogStepsParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters["LeapFrogSteps"]; }
    }
    public IFixedValueParameter<DoubleValue> LeapFrogStepSizeParameter {
      get { return (IFixedValueParameter<DoubleValue>)Parameters["LeapFrogStepSize"]; }
    }
    public IFixedValueParameter<DoubleValue> NoiseSigmaParameter {
      get { return (IFixedValueParameter<DoubleValue>)Parameters["NoiseSigma"]; }
    }

    public string ModelStructure {
      get { return ModelStructureParameter.Value.Value; }
      set { ModelStructureParameter.Value.Value = value; }
    }

    public int Iterations {
      get { return IterationsParameter.Value.Value; }
      set { IterationsParameter.Value.Value = value; }
    }

    public int Restarts {
      get { return RestartsParameter.Value.Value; }
      set { RestartsParameter.Value.Value = value; }
    }

    public int Seed {
      get { return SeedParameter.Value.Value; }
      set { SeedParameter.Value.Value = value; }
    }

    public bool SetSeedRandomly {
      get { return SetSeedRandomlyParameter.Value.Value; }
      set { SetSeedRandomlyParameter.Value.Value = value; }
    }

    public bool InitializeParametersRandomly {
      get { return InitParametersRandomlyParameter.Value.Value; }
      set { InitParametersRandomlyParameter.Value.Value = value; }
    }

    public int LeapFrogSteps {
      get { return LeapFrogStepsParameter.Value.Value; }
      set { LeapFrogStepsParameter.Value.Value = value; }
    }
    public double LeapFrogStepSize {
      get { return LeapFrogStepSizeParameter.Value.Value; }
      set { LeapFrogStepSizeParameter.Value.Value = value; }
    }
    public double NoiseSigma {
      get { return NoiseSigmaParameter.Value.Value; }
      set { NoiseSigmaParameter.Value.Value = value; }
    }

    [StorableConstructor]
    private BayesianNonlinearRegression(bool deserializing) : base(deserializing) { }
    private BayesianNonlinearRegression(BayesianNonlinearRegression original, Cloner cloner)
      : base(original, cloner) {
    }
    public BayesianNonlinearRegression()
      : base() {
      Problem = new RegressionProblem();
      Parameters.Add(new FixedValueParameter<StringValue>(ModelStructureParameterName, "The function for which the parameters must be fit (only numeric constants are tuned).", new StringValue("1.0 * x*x + 0.0")));
      Parameters.Add(new FixedValueParameter<IntValue>(IterationsParameterName, "The maximum number of iterations for constants optimization.", new IntValue(200)));
      Parameters.Add(new FixedValueParameter<IntValue>(RestartsParameterName, "The number of independent random restarts (>0)", new IntValue(10)));
      Parameters.Add(new FixedValueParameter<IntValue>(SeedParameterName, "The PRNG seed value.", new IntValue()));
      Parameters.Add(new FixedValueParameter<BoolValue>(SetSeedRandomlyParameterName, "Switch to determine if the random number seed should be initialized randomly.", new BoolValue(true)));
      Parameters.Add(new FixedValueParameter<BoolValue>(InitParamsRandomlyParameterName, "Switch to determine if the real-valued model parameters should be initialized randomly in each restart.", new BoolValue(false)));

      Parameters.Add(new FixedValueParameter<IntValue>("LeapFrogSteps", "LeapFrogSteps", new IntValue(10)));
      Parameters.Add(new FixedValueParameter<DoubleValue>("LeapFrogStepSize", "LeapFrogStepSize", new DoubleValue(0.1)));
      Parameters.Add(new FixedValueParameter<DoubleValue>("NoiseSigma", "NoiseSigma", new DoubleValue(0.1)));

      SetParameterHiddenState();

      InitParametersRandomlyParameter.Value.ValueChanged += (sender, args) => {
        SetParameterHiddenState();
      };
    }

    private void SetParameterHiddenState() {
      var hide = !InitializeParametersRandomly;
      RestartsParameter.Hidden = hide;
      SeedParameter.Hidden = hide;
      SetSeedRandomlyParameter.Hidden = hide;
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new BayesianNonlinearRegression(this, cloner);
    }

    #region nonlinear regression
    protected override void Run(CancellationToken cancellationToken) {
      IRegressionSolution bestSolution = null;
      if (SetSeedRandomly) Seed = (new System.Random()).Next();
      var rand = new MersenneTwister((uint)Seed);

      double[][] chain;
      if (InitializeParametersRandomly) {
        var qualityTable = new DataTable("RMSE table");
        qualityTable.VisualProperties.YAxisLogScale = true;
        var trainRMSERow = new DataRow("RMSE (train)");
        trainRMSERow.VisualProperties.ChartType = DataRowVisualProperties.DataRowChartType.Points;
        var testRMSERow = new DataRow("RMSE test");
        testRMSERow.VisualProperties.ChartType = DataRowVisualProperties.DataRowChartType.Points;

        qualityTable.Rows.Add(trainRMSERow);
        qualityTable.Rows.Add(testRMSERow);
        Results.Add(new Result(qualityTable.Name, qualityTable.Name + " for all restarts", qualityTable));
        bestSolution = CreateRegressionSolution(Problem.ProblemData, ModelStructure, Iterations, rand, LeapFrogSteps, LeapFrogStepSize, NoiseSigma, out chain);
        trainRMSERow.Values.Add(bestSolution.TrainingRootMeanSquaredError);
        testRMSERow.Values.Add(bestSolution.TestRootMeanSquaredError);
        for (int r = 0; r < Restarts; r++) {
          var solution = CreateRegressionSolution(Problem.ProblemData, ModelStructure, Iterations, rand, LeapFrogSteps, LeapFrogStepSize, NoiseSigma, out chain);
          trainRMSERow.Values.Add(solution.TrainingRootMeanSquaredError);
          testRMSERow.Values.Add(solution.TestRootMeanSquaredError);
          if (solution.TrainingRootMeanSquaredError < bestSolution.TrainingRootMeanSquaredError) {
            bestSolution = solution;
          }
        }
      } else {
        bestSolution = CreateRegressionSolution(Problem.ProblemData, ModelStructure, Iterations, rand, LeapFrogSteps, LeapFrogStepSize, NoiseSigma, out chain);
        var nRows = chain.First().Length;
        var chainTable = new DataTable("Chain");
        var rows = new DataRow[nRows];
        for (int i = 0; i < nRows; i++) {
          rows[i] = new DataRow(i.ToString());
        }
        foreach(var sample in chain) {
          for(int i=0;i<sample.Length;i++) {
            rows[i].Values.Add(sample[i]);
          }
        }
        for (int i = 0; i < nRows; i++) {
          chainTable.Rows.Add(rows[i]);
        }

        Results.Add(new Result("Chain", chainTable));
      }

      Results.Add(new Result(RegressionSolutionResultName, "The nonlinear regression solution.", bestSolution));
      Results.Add(new Result("Root mean square error (train)", "The root of the mean of squared errors of the regression solution on the training set.", new DoubleValue(bestSolution.TrainingRootMeanSquaredError)));
      Results.Add(new Result("Root mean square error (test)", "The root of the mean of squared errors of the regression solution on the test set.", new DoubleValue(bestSolution.TestRootMeanSquaredError)));
    }

    /// <summary>
    /// Detemines the posterior distribution for the model parameters using Hamiltonian Monte Carlo.
    /// Model is specified as infix expression containing variable names and numbers. 
    /// Prior distribution for the parameters is N(0,\lambda I)
    /// </summary>-
    /// <param name="problemData">Training and test data</param>
    /// <param name="modelStructure">The function as infix expression</param>
    /// <param name="maxIterations">Number of samples for HMC</param>
    public static IRegressionSolution CreateRegressionSolution(
      IRegressionProblemData problemData, string modelStructure, int maxIterations, IRandom random,
      int leapFrogSteps, double leapFrogStepSize, double noiseSigma, out double[][] chain
      ) {
      var parser = new InfixExpressionParser();
      var tree = parser.Parse(modelStructure);
      // parser handles double and string variables equally by creating a VariableTreeNode
      // post-process to replace VariableTreeNodes by FactorVariableTreeNodes for all string variables
      var factorSymbol = new FactorVariable();
      factorSymbol.VariableNames =
        problemData.AllowedInputVariables.Where(name => problemData.Dataset.VariableHasType<string>(name));
      factorSymbol.AllVariableNames = factorSymbol.VariableNames;
      factorSymbol.VariableValues =
        factorSymbol.VariableNames.Select(name =>
        new KeyValuePair<string, Dictionary<string, int>>(name,
        problemData.Dataset.GetReadOnlyStringValues(name).Distinct()
        .Select((n, i) => Tuple.Create(n, i))
        .ToDictionary(tup => tup.Item1, tup => tup.Item2)));

      foreach (var parent in tree.IterateNodesPrefix().ToArray()) {
        for (int i = 0; i < parent.SubtreeCount; i++) {
          var varChild = parent.GetSubtree(i) as VariableTreeNode;
          var factorVarChild = parent.GetSubtree(i) as FactorVariableTreeNode;
          if (varChild != null && factorSymbol.VariableNames.Contains(varChild.VariableName)) {
            parent.RemoveSubtree(i);
            var factorTreeNode = (FactorVariableTreeNode)factorSymbol.CreateTreeNode();
            factorTreeNode.VariableName = varChild.VariableName;
            factorTreeNode.Weights =
              factorTreeNode.Symbol.GetVariableValues(factorTreeNode.VariableName).Select(_ => 1.0).ToArray();
            // weight = 1.0 for each value
            parent.InsertSubtree(i, factorTreeNode);
          } else if (factorVarChild != null && factorSymbol.VariableNames.Contains(factorVarChild.VariableName)) {
            if (factorSymbol.GetVariableValues(factorVarChild.VariableName).Count() != factorVarChild.Weights.Length)
              throw new ArgumentException(
                string.Format("Factor variable {0} needs exactly {1} weights",
                factorVarChild.VariableName,
                factorSymbol.GetVariableValues(factorVarChild.VariableName).Count()));
            parent.RemoveSubtree(i);
            var factorTreeNode = (FactorVariableTreeNode)factorSymbol.CreateTreeNode();
            factorTreeNode.VariableName = factorVarChild.VariableName;
            factorTreeNode.Weights = factorVarChild.Weights;
            parent.InsertSubtree(i, factorTreeNode);
          }
        }
      }

      // TODO: useful?
      // initialize constants randomly
      // if (random != null) {
      //   foreach (var node in tree.IterateNodesPrefix().OfType<ConstantTreeNode>()) {
      //     double f = Math.Exp(NormalDistributedRandom.NextDouble(random, 0, 1));
      //     double s = random.NextDouble() < 0.5 ? -1 : 1;
      //     node.Value = s * node.Value * f;
      //   }
      // }

      double[] initialConstants;
      var negLogLikelihood = CreateNegLogLikelihoodFunction(problemData, tree, noiseSigma, out initialConstants);


      // create parameter sample
      var sampledParameters = HamiltonianMonteCarlo.SampleChain(initialConstants, negLogLikelihood, random, leapFrogStepSize, leapFrogSteps)
        .Take(maxIterations);

      chain = sampledParameters.ToArray();

      var model = new BayesianNonlinearRegressionModel(tree, chain,
        new SymbolicDataAnalysisExpressionTreeLinearInterpreter(), problemData.TargetVariable, problemData.AllowedInputVariables);
      var solution = model.CreateRegressionSolution((IRegressionProblemData)problemData.Clone());

      solution.Model.Name = "Regression Model";
      solution.Name = "Regression Solution";
      return solution;
    }

    private static Func<double[], Tuple<double, double[]>> CreateNegLogLikelihoodFunction(
      IRegressionProblemData problemData, ISymbolicExpressionTree tree, double noiseSigma, 
      out double[] initialConstants) {
      List<TreeToAutoDiffTermConverter.DataForVariable> parameters;
      TreeToAutoDiffTermConverter.ParametricFunction func;
      TreeToAutoDiffTermConverter.ParametricFunctionGradient funcGrad;
      TreeToAutoDiffTermConverter.TryConvertToAutoDiff(tree, false, false, out parameters, out initialConstants, out func, out funcGrad);

      double variance = noiseSigma * noiseSigma;

      IDataset ds = problemData.Dataset;
      var rows = problemData.TrainingIndices;
      int N = rows.Count();
      var xs = new double[N][];
      int row = 0;
      foreach (var r in rows) {
        int col = 0;
        xs[row] = new double[parameters.Count];
        foreach (var info in parameters) {
          if (ds.VariableHasType<double>(info.variableName)) {
            xs[row][col] = ds.GetDoubleValue(info.variableName, r + info.lag);
          } else if (ds.VariableHasType<string>(info.variableName)) {
            xs[row][col] = ds.GetStringValue(info.variableName, r) == info.variableValue ? 1 : 0;
          } else throw new InvalidProgramException("found a variable of unknown type");
          col++;
        }
        row++;
      }
      var ys = ds.GetDoubleValues(problemData.TargetVariable, rows).ToArray();

      return (double[] p) => {
        var logProbSum = 0.0;
        var scalingFactor = 1.0 / (2.0 * variance * N);
        double[] gSum = new double[p.Length];
        for (int i = 0; i < N; i++) {
          var fg = funcGrad(p, xs[i]);
          // sum up err
          var err = fg.Item2 - ys[i];
          logProbSum += scalingFactor * err * err;

//          var llik = (err * err / (2 * sigma * sigma))

          // sum up grad
          for (int j = 0; j < gSum.Length; j++) gSum[j] += scalingFactor * 2 * err * fg.Item1[j];
        }


        double f = logProbSum /* + N / 2.0 * Math.Log(variance) + N / 2.0 * Math.Log(2 * Math.PI) (constant factors) */;
        double[] g = gSum /* + N / 2.0 * Math.Log(variance) + N / 2.0 * Math.Log(2 * Math.PI) (constant factors) */;
        return Tuple.Create(f, g);
      };
    }


    #endregion
  }
}
